@@ -1,4 +1,4 @@
-/* Copyright (c) 2010-2014 Benjamin Dobell, Glass Echidna
+/* Copyright (c) 2010-2017 Benjamin Dobell, Glass Echidna
  
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
@@ -19,7 +19,8 @@
  THE SOFTWARE.*/
 
 // C Standard Library
-#include <stdio.h>
+#include <cstdio>
+#include <fstream>
 
 // libusb
 #include <libusb.h>
@@ -53,7 +54,9 @@
 #include "SessionSetupResponse.h"
 
 // Future versions of libusb will use usb_interface instead of interface.
+#ifndef usb_interface
 #define usb_interface interface
+#endif
 
 #define USB_CLASS_CDC_DATA 0x0A
 
@@ -66,11 +69,6 @@ const DeviceIdentifier BridgeManager::supportedDevices[BridgeManager::kSupported
 	DeviceIdentifier(BridgeManager::kVidSamsung, BridgeManager::kPidDroidCharge)
 };
 
-enum
-{
-	kDumpBufferSize = 4096,
-};
-
 enum
 {
 	kFileTransferSequenceMaxLengthDefault = 800,
@@ -78,31 +76,22 @@ enum
 	kFileTransferSequenceTimeoutDefault = 30000 // 30 seconds
 };
 
-enum
-{
-	kReceivePacketMaxAttempts = 5
-};
-
-enum
-{
-	kPitSizeMultiplicand = 4096
-};
-
 int BridgeManager::FindDeviceInterface(void)
 {
 	Interface::Print("Detecting device...\n");
 
 	struct libusb_device **devices;
-	int deviceCount = libusb_get_device_list(libusbContext, &devices);
+	unsigned int deviceCount = libusb_get_device_list(libusbContext, &devices);
 
-	for (int deviceIndex = 0; deviceIndex < deviceCount; deviceIndex++)
+	for (unsigned int deviceIndex = 0; deviceIndex < deviceCount; deviceIndex++)
 	{
 		libusb_device_descriptor descriptor;
 		libusb_get_device_descriptor(devices[deviceIndex], &descriptor);
 
 		for (int i = 0; i < BridgeManager::kSupportedDeviceCount; i++)
 		{
-			if (descriptor.idVendor == supportedDevices[i].vendorId && descriptor.idProduct == supportedDevices[i].productId)
+			if (descriptor.idVendor == supportedDevices[i].vendorId &&
+			    descriptor.idProduct == supportedDevices[i].productId)
 			{
 				heimdallDevice = devices[deviceIndex];
 				libusb_ref_device(heimdallDevice);
@@ -272,9 +261,16 @@ bool BridgeManager::ClaimDeviceInterface(void)
 
 bool BridgeManager::SetupDeviceInterface(void)
 {
+	// if altSettingIndex is 0 there should be no need
+	// to set alt_setting(?)
+	if (altSettingIndex == 0)
+		return (true);
+
 	Interface::Print("Setting up interface...\n");
 
-	int result = libusb_set_interface_alt_setting(deviceHandle, interfaceIndex, altSettingIndex);
+	int result = libusb_set_interface_alt_setting(deviceHandle,
+						      interfaceIndex,
+						      altSettingIndex);
 
 	if (result != LIBUSB_SUCCESS)
 	{
@@ -316,9 +312,20 @@ bool BridgeManager::InitialiseProtocol(void)
 	memcpy(dataBuffer, "ODIN", 4);
 	memset(dataBuffer + 4, 0, 1);
 
+#ifdef OS_LINUX
+	if (IsUbuntu())
+	{
+		Interface::Print("Resetting device...\n");
+		if (libusb_reset_device(deviceHandle))
+		{
+			Interface::PrintError("Failed to reset device!\n");
+		}
+	}
+#endif
+
 	if (!SendBulkTransfer(dataBuffer, 4, 1000))
 	{
-		Interface::PrintError("Failed to send handshake!");
+		Interface::PrintError("Failed to send handshake!\n");
 	}
 
 	// Expect "LOKE"
@@ -408,29 +415,8 @@ bool BridgeManager::DetectDevice(void)
 		return (false);
 	}
 
-	// Setup libusb log level.
-	switch (usbLogLevel)
-	{
-		case UsbLogLevel::None:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_NONE);
-			break;
-
-		case UsbLogLevel::Error:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_ERROR);
-			break;
-
-		case UsbLogLevel::Warning:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_WARNING);
-			break;
-
-		case UsbLogLevel::Info:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_INFO);
-			break;
-
-		case UsbLogLevel::Debug:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_DEBUG);
-			break;
-	}
+	// Set libusb log level.
+	SetUsbLogLevel(usbLogLevel);
 
 	// Get handle to Galaxy S device
 	struct libusb_device **devices;
@@ -443,7 +429,8 @@ bool BridgeManager::DetectDevice(void)
 
 		for (int i = 0; i < BridgeManager::kSupportedDeviceCount; i++)
 		{
-			if (descriptor.idVendor == supportedDevices[i].vendorId && descriptor.idProduct == supportedDevices[i].productId)
+			if (descriptor.idVendor == supportedDevices[i].vendorId &&
+			    descriptor.idProduct == supportedDevices[i].productId)
 			{
 				libusb_free_device_list(devices, deviceCount);
 
@@ -469,34 +456,13 @@ int BridgeManager::Initialise(bool resume)
 	if (result != LIBUSB_SUCCESS)
 	{
 		Interface::PrintError("Failed to initialise libusb. libusb error: %d\n", result);
-		Interface::Print("Failed to connect to device!");
+		Interface::Print("Failed to connect to device!\n");
 		return (BridgeManager::kInitialiseFailed);
 	}
 
 	// Setup libusb log level.
-	switch (usbLogLevel)
-	{
-		case UsbLogLevel::None:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_NONE);
-			break;
+	SetUsbLogLevel(usbLogLevel);
 
-		case UsbLogLevel::Error:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_ERROR);
-			break;
-
-		case UsbLogLevel::Warning:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_WARNING);
-			break;
-
-		case UsbLogLevel::Info:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_INFO);
-			break;
-
-		case UsbLogLevel::Debug:
-			libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_DEBUG);
-			break;
-	}
-	
 	result = FindDeviceInterface();
 
 	if (result != BridgeManager::kInitialiseSucceeded)
@@ -661,13 +627,61 @@ bool BridgeManager::SendBulkTransfer(unsigned char *data, int length, int timeou
 	return (result == LIBUSB_SUCCESS && dataTransferred == length);
 }
 
-bool BridgeManager::SendPacket(OutboundPacket *packet, int timeout, int sendEmptyTransferFlags) const
+int BridgeManager::ReceiveBulkTransfer(unsigned char *data, int length, int timeout, bool retry) const
+{
+	if (data == nullptr)
+	{
+		// HACK: It seems WinUSB ignores us when we try to read with length zero.
+		static unsigned char dummyData;
+		data = &dummyData;
+		length = 1;
+	}
+
+	int dataTransferred;
+	int result = libusb_bulk_transfer(deviceHandle, inEndpoint, data, length, &dataTransferred, timeout);
+
+	if (result != LIBUSB_SUCCESS && retry)
+	{
+		static const int retryDelay = 250;
+
+		if (verbose)
+			Interface::PrintError("libusb error %d whilst receiving bulk transfer.", result);
+
+		// Retry
+		for (int i = 0; i < 5; i++)
+		{
+			if (verbose)
+				Interface::PrintErrorSameLine(" Retrying...\n");
+
+			// Wait longer each retry
+			Sleep(retryDelay * (i + 1));
+
+			result = libusb_bulk_transfer(deviceHandle, inEndpoint, data, length, &dataTransferred, timeout);
+
+			if (result == LIBUSB_SUCCESS)
+				break;
+
+			if (verbose)
+				Interface::PrintError("libusb error %d whilst receiving bulk transfer.", result);
+		}
+
+		if (verbose)
+			Interface::PrintErrorSameLine("\n");
+	}
+
+	if (result != LIBUSB_SUCCESS)
+		return (result);
+
+	return (dataTransferred);
+}
+
+bool BridgeManager::SendPacket(OutboundPacket *packet, int timeout, int emptyTransferFlags) const
 {
 	packet->Pack();
 
-	if (sendEmptyTransferFlags & kSendEmptyTransferBefore)
+	if (emptyTransferFlags & kEmptyTransferBefore)
 	{
-		if (!SendBulkTransfer(nullptr, 0, kDefaultTimeoutSendEmptyTransfer, false) && verbose)
+		if (!SendBulkTransfer(nullptr, 0, kDefaultTimeoutEmptyTransfer, false) && verbose)
 		{
 			Interface::PrintWarning("Empty bulk transfer before sending packet failed. Continuing anyway...\n");
 		}
@@ -676,9 +690,9 @@ bool BridgeManager::SendPacket(OutboundPacket *packet, int timeout, int sendEmpt
 	if (!SendBulkTransfer(packet->GetData(), packet->GetSize(), timeout))
 		return (false);
 
-	if (sendEmptyTransferFlags & kSendEmptyTransferAfter)
+	if (emptyTransferFlags & kEmptyTransferAfter)
 	{
-		if (!SendBulkTransfer(nullptr, 0, kDefaultTimeoutSendEmptyTransfer, false) && verbose)
+		if (!SendBulkTransfer(nullptr, 0, kDefaultTimeoutEmptyTransfer, false) && verbose)
 		{
 			Interface::PrintWarning("Empty bulk transfer after sending packet failed. Continuing anyway...\n");
 		}
@@ -687,59 +701,44 @@ bool BridgeManager::SendPacket(OutboundPacket *packet, int timeout, int sendEmpt
 	return (true);
 }
 
-bool BridgeManager::ReceivePacket(InboundPacket *packet, int timeout) const
+bool BridgeManager::ReceivePacket(InboundPacket *packet, int timeout, int emptyTransferFlags) const
 {
-	unsigned char *buffer = packet->GetData();
-	unsigned int bufferSize = packet->GetSize();
-
-	int dataTransferred;
-	int result;
-
-	unsigned int attempt = 0;
-
-	static const int retryDelay = 250;
-
-	for (; attempt < kReceivePacketMaxAttempts; attempt++)
+	if (emptyTransferFlags & kEmptyTransferBefore)
 	{
-		if (attempt > 0)
+		if (ReceiveBulkTransfer(nullptr, 0, kDefaultTimeoutEmptyTransfer, false) < 0 && verbose)
 		{
-			if (verbose)
-				Interface::PrintErrorSameLine(" Retrying...\n");
-			
-			// Wait longer each retry
-			Sleep(retryDelay * (attempt + 1));
+			Interface::PrintWarning("Empty bulk transfer before receiving packet failed. Continuing anyway...\n");
 		}
-
-		result = libusb_bulk_transfer(deviceHandle, inEndpoint, buffer, bufferSize, &dataTransferred, timeout);
-
-		if (result >= 0)
-			break;
-
-		if (verbose)
-			Interface::PrintError("libusb error %d whilst receiving packet.", result);
 	}
 
-	if (verbose && attempt > 0)
-		Interface::PrintErrorSameLine("\n");
+	int receivedSize = ReceiveBulkTransfer(packet->GetData(), packet->GetSize(), timeout);
 
-	if (attempt == kReceivePacketMaxAttempts)
+	if (receivedSize < 0)
 		return (false);
 
-	if (dataTransferred != packet->GetSize() && !packet->IsSizeVariable())
+	if (static_cast<unsigned int>(receivedSize) != packet->GetSize() && !packet->IsSizeVariable())
 	{
 		if (verbose)
-			Interface::PrintError("Incorrect packet size received - expected size = %d, received size = %d.\n", packet->GetSize(), dataTransferred);
+			Interface::PrintError("Incorrect packet size received - expected size = %d, received size = %d.\n", packet->GetSize(), receivedSize);
 
 		return (false);
 	}
 
-	packet->SetReceivedSize(dataTransferred);
+	packet->SetReceivedSize(receivedSize);
 
 	bool unpacked = packet->Unpack();
 
 	if (!unpacked && verbose)
 		Interface::PrintError("Failed to unpack received packet.\n");
 
+	if (emptyTransferFlags & kEmptyTransferAfter)
+	{
+		if (ReceiveBulkTransfer(nullptr, 0, kDefaultTimeoutEmptyTransfer, false) < 0 && verbose)
+		{
+			Interface::PrintWarning("Empty bulk transfer after receiving packet failed. Continuing anyway...\n");
+		}
+	}
+
 	return (unpacked);
 }
 
@@ -903,7 +902,6 @@ int BridgeManager::ReceivePitFile(unsigned char **pitBuffer) const
 	unsigned char *buffer = new unsigned char[fileSize];
 	int offset = 0;
 
-	// NOTE: The PIT file appears to always be padded out to exactly 4 kilobytes.
 	for (unsigned int i = 0; i < transferCount; i++)
 	{
 		DumpPartPitFilePacket *requestPacket = new DumpPartPitFilePacket(i);
@@ -916,10 +914,12 @@ int BridgeManager::ReceivePitFile(unsigned char **pitBuffer) const
 			delete [] buffer;
 			return (0);
 		}
-		
+
+		int receiveEmptyTransferFlags = (i == transferCount - 1) ? kEmptyTransferAfter : kEmptyTransferNone;
+
 		ReceiveFilePartPacket *receiveFilePartPacket = new ReceiveFilePartPacket();
-		success = ReceivePacket(receiveFilePartPacket);
-		
+		success = ReceivePacket(receiveFilePartPacket, kDefaultTimeoutReceive, receiveEmptyTransferFlags);
+
 		if (!success)
 		{
 			Interface::PrintError("Failed to receive PIT file part #%d!\n", i);
@@ -1003,7 +1003,7 @@ bool BridgeManager::SendFile(FILE *file, unsigned int destination, unsigned int
 	}
 
 	FileSeek(file, 0, SEEK_END);
-	unsigned int fileSize = (unsigned int)FileTell(file);
+	unsigned long fileSize = (unsigned long)FileTell(file);
 	FileRewind(file);
 
 	ResponsePacket *fileTransferResponse = new ResponsePacket(ResponsePacket::kResponseTypeFileTransfer);
@@ -1024,14 +1024,14 @@ bool BridgeManager::SendFile(FILE *file, unsigned int destination, unsigned int
 	{
 		sequenceCount++;
 
-		int lastSequenceBytes = fileSize % (fileTransferSequenceMaxLength * fileTransferPacketSize);
+		unsigned int lastSequenceBytes = fileSize % (fileTransferSequenceMaxLength * fileTransferPacketSize);
 		lastSequenceSize = lastSequenceBytes / fileTransferPacketSize;
 
 		if (partialPacketByteCount != 0)
 			lastSequenceSize++;
 	}
 
-	unsigned int bytesTransferred = 0;
+	unsigned long bytesTransferred = 0;
 	unsigned int currentPercent;
 	unsigned int previousPercent = 0;
 	Interface::Print("0%%");
@@ -1067,7 +1067,7 @@ bool BridgeManager::SendFile(FILE *file, unsigned int destination, unsigned int
 		for (unsigned int filePartIndex = 0; filePartIndex < sequenceSize; filePartIndex++)
 		{
 			// NOTE: This empty transfer thing is entirely ridiculous, but sadly it seems to be required.
-			int sendEmptyTransferFlags = (filePartIndex == 0) ? kSendEmptyTransferNone : kSendEmptyTransferBefore;
+			int sendEmptyTransferFlags = (filePartIndex == 0) ? kEmptyTransferNone : kEmptyTransferBefore;
 
 			// Send
 			SendFilePartPacket *sendFilePartPacket = new SendFilePartPacket(file, fileTransferPacketSize);
@@ -1171,7 +1171,7 @@ bool BridgeManager::SendFile(FILE *file, unsigned int destination, unsigned int
 		{
 			EndPhoneFileTransferPacket *endPhoneFileTransferPacket = new EndPhoneFileTransferPacket(sequenceEffectiveByteCount, 0, deviceType, fileIdentifier, isLastSequence);
 
-			success = SendPacket(endPhoneFileTransferPacket, kDefaultTimeoutSend, kSendEmptyTransferBeforeAndAfter);
+			success = SendPacket(endPhoneFileTransferPacket, kDefaultTimeoutSend, kEmptyTransferBeforeAndAfter);
 			delete endPhoneFileTransferPacket;
 
 			if (!success)
@@ -1185,7 +1185,7 @@ bool BridgeManager::SendFile(FILE *file, unsigned int destination, unsigned int
 		{
 			EndModemFileTransferPacket *endModemFileTransferPacket = new EndModemFileTransferPacket(sequenceEffectiveByteCount, 0, deviceType, isLastSequence);
 
-			success = SendPacket(endModemFileTransferPacket, kDefaultTimeoutSend, kSendEmptyTransferBeforeAndAfter);
+			success = SendPacket(endModemFileTransferPacket, kDefaultTimeoutSend, kEmptyTransferBeforeAndAfter);
 			delete endModemFileTransferPacket;
 
 			if (!success)
@@ -1223,24 +1223,53 @@ void BridgeManager::SetUsbLogLevel(UsbLogLevel usbLogLevel)
 		switch (usbLogLevel)
 		{
 			case UsbLogLevel::None:
-				libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_NONE);
+				libusb_set_option(libusbContext, LIBUSB_OPTION_LOG_LEVEL, LIBUSB_LOG_LEVEL_NONE);
 				break;
 
 			case UsbLogLevel::Error:
-				libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_ERROR);
+				libusb_set_option(libusbContext, LIBUSB_OPTION_LOG_LEVEL, LIBUSB_LOG_LEVEL_ERROR);
 				break;
 
 			case UsbLogLevel::Warning:
-				libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_WARNING);
+				libusb_set_option(libusbContext, LIBUSB_OPTION_LOG_LEVEL, LIBUSB_LOG_LEVEL_WARNING);
 				break;
 
 			case UsbLogLevel::Info:
-				libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_INFO);
+				libusb_set_option(libusbContext, LIBUSB_OPTION_LOG_LEVEL, LIBUSB_LOG_LEVEL_INFO);
 				break;
 
 			case UsbLogLevel::Debug:
-				libusb_set_debug(libusbContext, LIBUSB_LOG_LEVEL_DEBUG);
+				libusb_set_option(libusbContext, LIBUSB_OPTION_LOG_LEVEL, LIBUSB_LOG_LEVEL_DEBUG);
 				break;
 		}
 	}
 }
+
+#ifdef OS_LINUX
+bool BridgeManager::IsUbuntu()
+{
+	std::ifstream os_release("/etc/os-release");
+	std::string line, entry, os;
+	int pos;
+	while (std::getline(os_release, line))
+	{
+		pos = line.find("=");
+		entry = line.substr(0, pos);
+		if (entry == "ID")
+		{
+			os = line.substr(pos+1);
+			if (verbose)
+			{
+				Interface::Print("Linux distro ID: %s\n",
+						 os.c_str());
+			}
+			if (os == "ubuntu")
+			{
+				return true;
+			}
+			break;
+		}
+	}
+	return false;
+}
+#endif
